/* gqcf.c */

/* Krzysztof Zmudzinski
   CS 737 Class Project
   Controlled Transfer of Qclic Files
   Date of completion: 5/5/97

   The purpose of this project was to develop an algorithm for
   controlled transfer of Qclic files.  `Controlled' means that the
   user specifies what files should be transferred, how much bandwidth
   should be used to transfer these files, and how much time should
   the transfer take.  The algorithm is responsible for transferring
   the specified files without exceeding the specified bandwidth and time.

   This progam is only a model of a client-server scheme that uses pipes
   to communicate between the client and the server.  The final program
   should be implemented using Berkeley sockets which should be trivial
   given the current state of the model.  
   The only new issues that will have to be addressed are:
   1. passing arguments to the server (file names, bandwidth, time, etc.);
   2. the number of sockets (as opposed to the current number of pipes);

   There are two basic modes in which files are transferred:
   (note that transferSize = bandwidth * transferTime)
   
   1. alternate - a file is opened, one layer is transferred, the 
   file is closed, next file is opened, one layer is transferred, ..., 
   the first is opened again, ..., and so on until we reach transferSize;

   2. sequential - transferSize is equally divided between all files; 
   each file is opened only once and its share of transferSize is 
   transferred;

   Each of the two basic modes has three variations:
   1. exact - transferSize is really equal to bandwidth * transferTime;
   layers may be transferred partially;

   2. next scan - transferSize may be exceeded so layers are transferred
   in their entirety;

   3. previous scan - transferSize may not be exceeded and layers must be
   transferred in their entirety;
   
   The following are transfer modes which can be passed to GetQclicFiles:
   
   TM_ALT_EXACT - we alternate files and stop as soon as we reach
   transferSize regardless whether we read the entire layer of the 
   current file or not;

   TM_ALT_NEXT_SCAN - we alternate files and stop once we finish 
   reading the entire layer of the current file such the total 
   transferred size will be bigger or equal to transferSize;

   TM_ALT_PREV_SIZE - we alternate files and stop once we know that 
   reading the entire layer of the current file would cause the total 
   transferred size to be more than transferSize;

   TM_SEQ_EXACT - we divide transferSize equally among all files and 
   transfer each file up to its share of transferSize; layers may be
   transferred partially;

   TM_SEQ_NEXT_SCAN - we divide transferSize equally among all files 
   and transfer each file up to and including the layer that will use 
   the size allotted to that file and possibly (and likely) exceed it;

   TM_SEQ_PREV_SCAN - we divide transferSize equally among all files 
   and transfer each file up to but excluding the layer that would cause 
   the size allotted to that file to be exceeded;
   
   The above works fine for similar files (the same transferred size 
   corresponds to the same quality); for files that are different,
   we could look at the quality and not the size when making decisions
   how much of each file should be transferred; the assumption here is
   that similar-size files will be transferred and Qclic algorithm
   to compress them was run with similar parameters.  In case the 
   program was to be modified to include quality in transfer
   considerations, some hooks have been added.

   There are also two ways to control file transfer:
   1. time-out - after each send, we compare the time it should take to
   transfer the current total number of transferred bytes at the specified
   bandwidth to the actual time it took us to transfer those bytes and wait
   for the difference, if any;

   2. buffer - we change the size of buffer used to transfer files to
   control the transfer rate; this solution has its drawbacks for low
   bandwidth and large files -- more system calls have to be executed;
   
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/wait.h>
#include <errno.h>

#define MAX_FILE_NUM 256
#define MAX_FILE_NAME 256
#define MAX_LINE 80
#define MAX_SCANS 100
#define MAX_CTRL_BUFF 10
#define MILLION 1000000.0
#define MAX_BUFF 4096
#define MIN_BUFF 16

#define TM_ALT_EXACT 1
#define TM_ALT_NEXT_SCAN 2
#define TM_ALT_PREV_SCAN 3
#define TM_SEQ_EXACT 4
#define TM_SEQ_NEXT_SCAN 5
#define TM_SEQ_PREV_SCAN 6

#define TM_FIRST 1
#define TM_LAST 6

#define CM_TIMEOUT 1
#define CM_BUFFER 2

#define CM_FIRST 1
#define CM_LAST 2

#ifdef ALL_OUTPUT
#define INFO
#define CTRL
#define FILEOP
#define ACK
#define FILEIO
#define CLIENT
#define SERVER
#define WATOUT
#define WAIT
#endif

double minwait = 1000000, maxwait = 0, totalwait = 0;
int numofwaits = 0;
int minbuf = 0, maxbuf = 0;

/* ReadLine
   
   reads one line from file described by iFD into buffer sBuff;
   replaces '\n' with '\0';
   returns number of characters read or EOF
*/

int 
ReadLine(char sBuff[], 
	 int iFD)
{
  int iCount;
  char cChar;

  iCount = 0;
  
  while (read(iFD, &cChar, 1) == 1) {
    if (cChar == '\n') {
      sBuff[iCount] = '\0';
      return (iCount + 1);
    }
    sBuff[iCount] = cChar;
    iCount++;
  }
  
  sBuff[iCount] = '\0';
  return EOF;
}

/* ReadFileInfo

   opens jsdir file associated with fileName, reads information
   necessary to transfer fileName, and returns it in passed arguments;

   returns 0 if jsdir file exists and contains information for at
   last two layers; returns -1 otherwise;

   format of jsdir file:
   ---------------------
   scan offset <offset>
   scan offset <offset>
   ...
   end offset <offset>
   #if nplanes==1
   PSNR: <psnr> dB
   PSNR: <psnr> dB
   ..
   #elseif nplanes==3
   RPSNR: <rpsnr> dB GPSNR: <gpsnr> dB BPSNR: <bpsnr> dB
   or
   YPSNR: <rpsnr> dB CbPSNR: <gpsnr> dB CrPSNR: <bpsnr> dB
   ...
  
   only number of scans (layers) and corresponding offsets are used;
   psnr, gpsnr, and bpsnr may be used in future modifications;
*/

int 
ReadFileInfo(char *fileName,
	     int *nscansplus1,
	     int *offsets, 
	     double *psnr, 
	     double *gpsnr, 
	     double *bpsnr)
{

  char jsdirname[MAX_FILE_NAME];
  int fd, l, i, from;
  char buff[MAX_LINE];
  char foo1[20], foo2[20], foo3[20], foo4[20], foo5[20], foo6[20];
  
  sprintf(jsdirname, "%s.jsdir", fileName);
  if ((fd = open(jsdirname, O_RDONLY)) < 0) {
    fprintf(stderr, "server error: could not open %s\n", jsdirname);
    fflush(stderr);
    return -1;
  }
      
  *nscansplus1 = 0;
  l = ReadLine(buff, fd);
  while ((l != EOF) && (strncmp(buff, "end", 3))) {
    sscanf(buff, "%s%s%d", foo1, foo2, &offsets[*nscansplus1]);
    (*nscansplus1)++;
    l = ReadLine(buff, fd);
  }
  if (l == EOF) {
    close(fd);
    return -1;
  }
  sscanf(buff, "%s%s%d", foo1, foo2, &offsets[*nscansplus1]);
  (*nscansplus1)++;

  /* if only header or incomplete information present */
  if (*nscansplus1 < 2) {
    return -1;
  }
  
  i = 1;
  l = ReadLine(buff, fd);
  from = 0;
  
  while ((l != EOF) && (i < *nscansplus1)) {
    if (!strncmp(buff, "RPSNR", 5)) {
      sscanf(buff, "%s%lf%s%s%lf%s%s%lf%s", 
	     foo1, &psnr[i], foo2, foo3, &gpsnr[i], 
	     foo4, foo5, &bpsnr[i], foo6);

#ifdef INFO
      fprintf(stdout, "server: size:%d B, R:%f dB, G:%f dB, B:%f dB\n",
	      offsets[i] - from, psnr[i], gpsnr[i], bpsnr[i]);
      fflush(stdout);
#endif

    } else if (!strncmp(buff,"YPSNR", 5)) {
      sscanf(buff, "%s%lf%s%s%lf%s%s%lf%s", 
	     foo1, &psnr[i], foo2, foo3, &gpsnr[i], 
	     foo4, foo5, &bpsnr[i], foo6);

#ifdef INFO
      fprintf(stdout, "server: size%d B, Y:%f dB, Cb:%f dB, Cr:%f dB\n",
	      offsets[i] - from, psnr[i], gpsnr[i], bpsnr[i]);
      fflush(stdout);
#endif

    } else {
      sscanf(buff, "%s%lf%s", foo1, &psnr[i], foo2);

#ifdef INFO
      fprintf(stdout, "server: size:%d B, Gray:%f dB\n", 
	      offsets[i] - from, psnr[i]);
      fflush(stdout);
#endif

    }
    from = offsets[i];
    i++;
    l = ReadLine(buff, fd);
  }

  close(fd);
  
#ifdef INFO
  if (i < *nscansplus1)  {
    while (i < *nscansplus1) {
      fprintf(stdout, "server: size:%d B\n", offsets[i] - from);
      fflush(stdout);
      from = offsets[i];
      i++;
    }
    fprintf(stdout, "END\n");
    fflush(stdout);
  }

  fprintf(stdout, "END\n");
  fflush(stdout);
#endif

  return 0;
  
}

/* ReceiveCtrl
   
   Client calls this when select indicates that control pipe has some data;
   there are three messages that may be sent from Server to Client through
   the control pipe:
   1. START - indicates that the next chunk of data arriving through the 
   data pipe is part of the next file in the target file list;
   2. STOP - indicates that the last chunk of data sent through the data
   pipe was the last part of the current traget file; file should be closed;
   3. QUIT - all files were transferred; no more data to transfer;

   sends an ACK for each type of message received; 
   returns -1 if there are errors encountered; 1 if QUIT messagewas received;
   0 otherwise (no errors);
*/

int
ReceiveCtrl(int pipeCtrl, 
	    int pipeAck,
	    char targetFiles[][MAX_FILE_NAME],
	    int *fd)
{

  static int fn = 0;
  /* first time around when START is received, 
     files will be created; otherwise, they will be opened */
  static int createFiles = 1; 
  char buff[MAX_CTRL_BUFF];

  (void) ReadLine(buff, pipeCtrl);
  
#ifdef CTRL
  fprintf(stdout, "client: control received %s\n", buff);
  fflush(stdout);
#endif
  
  if (!strncmp(buff, "STOP", 4)) {
    if (*fd != -1) {
      if (close(*fd) < 0) {
	fprintf(stderr, "client error: can't close file\n");
	fflush(stderr);
      } else {
	
#ifdef FILEOP
	fprintf(stdout, "client: closed file %d\n", *fd);
	fflush(stdout);
#endif
	
      }
    }
  }
  
  if (!strncmp(buff, "START", 5)) {
    if (createFiles) {
      if ((*fd = creat(targetFiles[fn], 438)) < 0) {
	fprintf(stderr, "client error: can't create file %s\n",
		targetFiles[fn]);
	fflush(stderr);
	return -1;
      } else {
	
#ifdef FILEOP
	fprintf(stdout, "client: created file %s %d\n", targetFiles[fn], *fd);
	fflush(stdout);
#endif
	
      }
    } else {
      if ((*fd = open(targetFiles[fn], O_WRONLY | O_APPEND)) < 0) {
	fprintf(stderr, "client error: can'topen file %s\n", targetFiles[fn]);
	fflush(stderr);
	return -1;
      } else {
	
#ifdef FILEOP
	fprintf(stdout, "client: opened file %s %d\n", targetFiles[fn], *fd);
	fflush(stdout);
#endif
	
      }
    }
    fn++;
    if (targetFiles[fn][0] == '\0') {
      createFiles = 0;
      fn = 0;
    }
  }
  
  if (write(pipeAck, "ACK\n", 4) < 0) {
    fprintf(stderr, "client error: error writing control\n");
    fflush(stderr);
    return -1;
  }
  
#ifdef ACK
  fprintf(stdout, "client: sent ctrl ack\n");
  fflush(stdout);
#endif
  
  if (!strncmp(buff, "QUIT", 4)) {
    return 1;
  }

  return 0;

}

/* ReceiveData
   
   Client calls this when select indicates that data pipe has some data;
   data is read and written to the current target file;
   sends an ACK;
   returns -1 if there are error encountered; 0 otherwise (no errors);
*/

int
ReceiveData(int pipeData,
	    int pipeAck,
	    int bufferSize,
	    int fd)
{
  int n;
  char buff[bufferSize];

  if ((n = read(pipeData, buff, bufferSize)) > 0) {
    if (write(fd, buff, n) < 0) {
      fprintf(stderr, "client error: error writing file %d %d\n", fd, errno);
      fflush(stderr);
      return -1;
    }
    
#ifdef FILEIO
    fprintf(stdout, "client: %d bytes of data received\n", n);
    fflush(stdout);
#endif
    
  } else if (n < 0) {
    fprintf(stderr, "client error: error reading data\n");
    fflush(stderr);
    return -1;
  }
  
  if (write(pipeAck, "ACK\n", 4) < 0) {
    fprintf(stderr, "client error: error writing control\n");
    fflush(stderr);
    return -1;
  }
  
#ifdef ACK
  fprintf(stdout, "client: sent DATA ack\n");
  fflush(stdout);
#endif

  return 0;

}

void 
Client(char targetFiles[][MAX_FILE_NAME],
       int pipeData, 
       int pipeCtrl, 
       int pipeAck,
       long transferTime, 
       int bufferSize) 
{
  struct timeval twait;
  int fd, n;
  fd_set fs;
  int maxfdp1;

  fd = -1; /* assume no file was opened */
  twait.tv_sec = transferTime;

  while (1) {

    /* initialize arguments for select */
    FD_ZERO(&fs);
    FD_SET(pipeData, &fs);
    FD_SET(pipeCtrl, &fs);

    if (pipeData > pipeCtrl) {
      maxfdp1 = pipeData;
    } else {
      maxfdp1 = pipeCtrl;
    }

    maxfdp1++;

    if ((n = select(maxfdp1, &fs, 
		    (fd_set *) 0, (fd_set *) 0, &twait) < 0)) {
      fprintf(stderr, "client error: select error\n");
      fflush(stderr);
    }

    if (FD_ISSET(pipeCtrl, &fs)) {
      if (ReceiveCtrl(pipeCtrl, pipeAck, targetFiles,&fd) != 0) {
	break;
      }
      continue;
    }

    if (FD_ISSET(pipeData, &fs)) {
      if (ReceiveData(pipeData, pipeAck, bufferSize, fd) != 0) {
	break;
      }
      continue;
    }

    if (n == 0) {
      fprintf(stderr, "client error: timeout\n");
      fflush(stderr);
      break;
    }

  }
  
#ifdef CLIENT
  fprintf(stdout, "client: quiting...\n");
  fflush(stdout);
#endif

}

void
WaitAfter(long totalSize, 
	  long bandwidth,
	  int *bufferSize,
	  double usecStart,
	  int controlMode)
{
  struct timeval tcurr, twait;
  double usecCurr, usec;

  if (gettimeofday(&tcurr, (struct timezone *) 0) < 0) {
    return;
  } else {
    usecCurr = (double) tcurr.tv_sec * MILLION + (double) tcurr.tv_usec;
  }
  
  if (controlMode == CM_BUFFER) {
    *bufferSize = ((usecCurr - usecStart) * (double) bandwidth * 
		   (double) *bufferSize) / (MILLION * (double) totalSize);

    if (*bufferSize < MIN_BUFF) {
      *bufferSize = MIN_BUFF;
    } else if (*bufferSize > MAX_BUFF) {
      *bufferSize = MAX_BUFF;
    }

#ifdef WATOUT
    fprintf(stdout, "wait: new buffer: %d\n", *bufferSize);
    fflush(stdout);
#endif

    if (*bufferSize < minbuf) {
      minbuf = *bufferSize;
    } else if (*bufferSize > maxbuf) {
      maxbuf = *bufferSize;
    }

  } else {

    usec = (MILLION * (double) totalSize) / (double) bandwidth 
      + usecStart - usecCurr;
    
#ifdef WATOUT
    fprintf(stdout, "wait: usec: %f\n", usec);
    fflush(stdout);
#endif
    
    if (usec < 0) {
      return;
    }
    if (usec < minwait) {
      minwait = usec;
    } else if (usec > maxwait) {
      maxwait = usec;
    }

    totalwait += usec;
    numofwaits++;
    twait.tv_sec = (long) (usec / MILLION);
    twait.tv_usec = (long) (usec - (double) twait.tv_sec * MILLION);
    
#ifdef WATOUT
    fprintf(stdout, "wait: twait.tv_sec: %ld; twait.tv_usec: %ld\n", 
	    twait.tv_sec, twait.tv_usec);
    fflush(stdout);
#endif
    
    if (select(0, 
	       (fd_set *) 0, 
	       (fd_set *) 0, 
	       (fd_set *) 0, 
	       &twait) < 0) {
      fprintf(stderr, "server error: select error\n");
      fflush(stderr);
    }
  }
}

int
TransferFile(char *fileName,
	     int from,
	     int nbytes,
	     int pipeData,
	     int pipeAck,
	     int *bufferSize, 
	     int *dataRead,
	     long *totalSize,
	     int *currSize,
	     int transferMode, 
	     long bandwidth,
	     long maxSize,
	     double usecStart,
	     int controlMode)
{
  char buff[MAX_BUFF];
  int nloops, remaining;
  int n, fd;
  
  if ((fd = open(fileName, O_RDONLY, 0)) < 0) {
    fprintf(stderr, "server error: could not open %s\n", fileName);
    fflush(stderr);
    return -1;
  }
  
  if (lseek(fd, from, SEEK_SET) < 0) {
    fprintf(stderr, "server error: seek error\n");
    fflush(stderr);
    close(fd);
    return -1;
  }

  nloops = nbytes / *bufferSize;
  remaining = nbytes % *bufferSize;
  
  while (nloops > 0) {
    if ((n = read(fd, buff, *bufferSize)) < 0) {
      fprintf(stderr, "server error: read error\n");
      fflush(stderr);
      close(fd);
      return -1;
    }
    
    *dataRead = 1;

#ifdef FILEIO
    fprintf(stdout, "server: %d bytes read\n", n);
    fflush(stdout);
#endif

    if ((n = write(pipeData, buff, *bufferSize)) < 0) {
      fprintf(stderr, "server error: write error\n");
      fflush(stderr);
      close(fd);
      return -1;
    }
    
    *totalSize += n;
    *currSize += n;

#ifdef FILEIO
    fprintf(stdout, "server: %d bytes written\n", n);
    fflush(stdout);
#endif

    (void) ReadLine(buff, pipeAck);

#ifdef ACK
    fprintf(stdout, "server: received data ack\n");
    fflush(stdout);
#endif

    if (transferMode == TM_ALT_EXACT || transferMode == TM_ALT_PREV_SCAN) {
      if (*totalSize >= maxSize) {
	break;
      }
    }

    if (transferMode == TM_SEQ_EXACT || transferMode == TM_SEQ_PREV_SCAN) {
      if (*currSize >= maxSize) {
	break;
      }
    }

#ifdef WAIT
    WaitAfter(*totalSize, bandwidth, bufferSize, usecStart, controlMode);
#endif

    if (controlMode == CM_TIMEOUT) {
      nloops--;
    } else {
      nbytes -= n;
      nloops = nbytes / *bufferSize;
      remaining = nbytes % *bufferSize;
    }
  
  }
  
  if (remaining < 1) {
    close(fd);
    return 0;
  }

  if ((*totalSize < maxSize || transferMode == TM_ALT_NEXT_SCAN) ||
      (*currSize < maxSize || transferMode == TM_SEQ_NEXT_SCAN)) {
    if ((n = read(fd, buff, remaining)) < 0) {
      fprintf(stderr, "server error: read error\n");
      fflush(stderr);
      close(fd);
      return -1;
    }
    
    *dataRead = 1;

#ifdef FILEIO
    fprintf(stdout, "server: %d bytes read\n", n);
    fflush(stdout);
#endif

    if ((n = write(pipeData, buff, remaining)) < 0) {
      fprintf(stderr, "server error: write error\n");
      fflush(stderr);
      return -1;
    }
      
    *totalSize += n;
    *currSize += n;

#ifdef FILEIO
    fprintf(stdout, "server: %d bytes written\n", n);
    fflush(stdout);
#endif

    (void) ReadLine(buff, pipeAck);

#ifdef ACK
    fprintf(stdout, "server: received data ack\n");
    fflush(stdout);
#endif

#ifdef WAIT
    WaitAfter(*totalSize, bandwidth, bufferSize, usecStart, controlMode);
#endif
  }

  close(fd);
  return 0;

}

int
SendCtrl(char *ctrlStr,
	 int pipeCtrl,
	 int pipeAck)
{

  char buff[MAX_CTRL_BUFF];

  sprintf(buff, "%s\n", ctrlStr);
  if (write(pipeCtrl, buff, strlen(buff)) != strlen(buff)) {
    fprintf(stderr, "server error: control write error\n");
    fflush(stderr);
    return -1;
  }
  
#ifdef CTRL
  fprintf(stdout, "server: %s sent\n", ctrlStr);
  fflush(stdout);
#endif
  
  if (ReadLine(buff, pipeAck) < 0) {
    fprintf(stderr, "server error: control read error\n");
    fflush(stderr);
    return -1;
  }
  
#ifdef ACK
  fprintf(stdout, "server: received %s ack\n", ctrlStr);
  fflush(stdout);
#endif

  return 0;

}

int
GetScanExtent(char *fileName,
	      int currScan,
	      int *from,
	      int *toread)
{
  int offsets[MAX_SCANS];
  int numscans;

  /* hook for quality */
  double psnr[MAX_SCANS], gpsnr[MAX_SCANS], bpsnr[MAX_SCANS];

  if (ReadFileInfo(fileName, &numscans, offsets, psnr, gpsnr, bpsnr) != 0) {
    fprintf(stderr, "server error: error reading %s.jsdir\n", fileName);
    fflush(stderr);
    return -1;
  }

  if (currScan > numscans - 1) {
    return 1;
  }
  
  /* there is an error in the original algorithm (found in jview.c):
     1. an image file starts at byte 0 instead of 1;
     2. the beginning of a scan is used as the size of the previous 
     scan which is 1 byte bigger than the actual scan size since the 
     file starts at 0;
     3. the end of the last scan is used as such 
  */
  
  if (currScan == 1) {
    *from = 1; /* 1. */
  } else {
    *from = offsets[currScan - 1];
  }
  
  *toread = offsets[currScan] - *from; /* 1. */
  /* last offset is the end of the offset, not the
     beginning of the next one (2. & 3.) */
  if (currScan == numscans - 1) {
    (*toread)++;
  }

#ifdef INFO
  fprintf(stdout, "server: scan # %d starts at %d and has %d bytes\n", 
	  currScan, *from, *toread);
  fflush(stdout);
#endif
      
  return 0;
  
}

int
TransferFileAlt(char sourceFiles[][MAX_FILE_NAME],
		int pipeData, 
		int pipeCtrl, 
		int pipeAck,
		long bandwidth, 
		long transferSize,
		int transferMode,
		int bufferSize,
		double usecStart,
		int controlMode)
{
  int fn; /* number of file from sourceFiles currently processed */
  int from, toread;
  long totalSize;
  int currSize, currScan;
  int dataRead;

  totalSize = 0;

  currScan = 0;
  /* assume that some data will be read */
  dataRead = 1;
  
  while ((totalSize < transferSize) && dataRead) {

    currScan++;
    currSize = 0; /* not really needed; required by TransferFile */
    
    fn = 0;
    /* assume that no data will be read */
    dataRead = 0;
    
    while (sourceFiles[fn][0] != '\0') {
      
      if (SendCtrl("START", pipeCtrl, pipeAck) != 0) {
	return totalSize;
      }
      
      if (GetScanExtent(sourceFiles[fn], currScan, &from, &toread) != 0) {
	fn++;
	continue;
      }

      if (totalSize + toread > transferSize) {
	if (transferMode == TM_ALT_EXACT) {
	  toread = transferSize - totalSize;
	} else if (transferMode == TM_ALT_PREV_SCAN) {
	  break;
	}
      }

      if (TransferFile(sourceFiles[fn],
		       from - 1, 
		       toread,
		       pipeData,
		       pipeAck,
		       &bufferSize, 
		       &dataRead,
		       &totalSize,
		       &currSize,
		       transferMode,
		       bandwidth,
		       transferSize,
		       usecStart,
		       controlMode) != 0) {
	fprintf(stderr, "server error: transfer file failed\n");
	fflush(stderr);
      }

      fn++;
      
      if (SendCtrl("STOP", pipeCtrl, pipeAck) != 0) {
	return totalSize;
      }
      
      if (totalSize >= transferSize) {
	break;
      }

    }
    
    if (transferMode == TM_ALT_PREV_SCAN) {
      if (totalSize + toread > transferSize) {
	break;
      }
    }
    
  }
  
  return totalSize;

}

int
TransferFileSeq(char sourceFiles[][MAX_FILE_NAME],
		int pipeData, 
		int pipeCtrl, 
		int pipeAck,
		long bandwidth, 
		long transferSize,
		int transferMode,
		int bufferSize,
		double usecStart,
		int controlMode)
{
  int numberOfFiles, fn;
  int from, toread;
  long shareSize, totalSize;
  int currSize, currScan;
  int dataRead;

  /* find number of files */
  fn = 0;
  while (sourceFiles[fn][0] != '\0') {
    fn++;
  }
  
  numberOfFiles = fn;
  shareSize = transferSize / numberOfFiles;

  totalSize = 0;

  for (fn = 0; fn < numberOfFiles; fn++) {
    
    if (SendCtrl("START", pipeCtrl, pipeAck) != 0) {
      return totalSize;
    }
    
    currScan = 0;
    currSize = 0;

    dataRead = 1;
    
    while ((currSize < shareSize) && dataRead) {
      
      currScan++;
      dataRead = 0;
      
      if (GetScanExtent(sourceFiles[fn], currScan, &from, &toread) != 0) {
	break;
      }
      
      if (currSize + toread > shareSize) {
	if (transferMode == TM_SEQ_EXACT) {
	  toread = shareSize - currSize;
	} else if (transferMode == TM_SEQ_PREV_SCAN) {
	  break;
	}
      }  

      if (TransferFile(sourceFiles[fn],
		       from - 1, 
		       toread,
		       pipeData,
		       pipeAck,
		       &bufferSize, 
		       &dataRead,
		       &totalSize,
		       &currSize,
		       transferMode,
		       bandwidth,
		       shareSize,
		       usecStart,
		       controlMode) != 0) {
	fprintf(stderr, "server error: transfer file failed\n");
	fflush(stderr);
      }
      
    }
    
    if (SendCtrl("STOP", pipeCtrl, pipeAck) != 0) {
      return totalSize;
    }

  }

  return totalSize;
  
}

void 
Server(char sourceFiles[][MAX_FILE_NAME],
       int pipeData, 
       int pipeCtrl, 
       int pipeAck,
       long bandwidth, 
       long transferTime, 
       int transferMode,
       int bufferSize,
       int controlMode)
{
  struct timeval tstart, tend;
  double usecStart, usecEnd;
  long transferSize;

  if (gettimeofday(&tstart, (struct timezone *) 0) < 0) {
    fprintf(stderr, "server error: gettimeofday failed\n");
  }

  usecStart = (double) tstart.tv_sec * MILLION + (double) tstart.tv_usec;

  transferSize = bandwidth * transferTime;

  switch (transferMode) {
  case TM_ALT_EXACT :
  case TM_ALT_NEXT_SCAN :
  case TM_ALT_PREV_SCAN :
    transferSize = TransferFileAlt(sourceFiles, pipeData, pipeCtrl, pipeAck, 
				   bandwidth, transferSize, transferMode, 
				   bufferSize, usecStart, controlMode);
    break;
  case TM_SEQ_EXACT :
  case TM_SEQ_NEXT_SCAN :
  case TM_SEQ_PREV_SCAN :
    transferSize = TransferFileSeq(sourceFiles, pipeData, pipeCtrl, pipeAck, 
				   bandwidth, transferSize, transferMode, 
				   bufferSize, usecStart, controlMode);
    break;
  }

  (void) SendCtrl("QUIT", pipeCtrl, pipeAck);

  if (gettimeofday(&tend, (struct timezone *) 0) < 0) {
    usecEnd = usecStart + transferTime;
  } else {
    usecEnd = (double) tend.tv_sec * MILLION + (double) tend.tv_usec;
  }

/*#ifdef SERVER*/
  fprintf(stdout, 
	  "server: transferred %ld bytes in %f sec at rate of %f Bs\n", 
	  transferSize, (usecEnd - usecStart) / MILLION, 
	  (double) transferSize / ((usecEnd - usecStart) / MILLION));
  fflush(stdout);
/*#endif*/

#ifdef SERVER
  fprintf(stdout, "server: quiting...\n");
  fflush(stdout);
#endif

  fprintf(stdout, "minwait:%f;maxwait:%f;totalwait:%f;numofwaits:%d\n", 
	  minwait, maxwait, totalwait, numofwaits);
  fprintf(stdout, "minbuf:%d;maxbuf:%d\n", minbuf, maxbuf);
  
}

void
GetQclicFiles(char sourceFiles[][MAX_FILE_NAME],
	      char targetFiles[][MAX_FILE_NAME],
	      long bandwidth, 
	      long transferTime, 
	      int transferMode,
	      int clientBufferSize,
	      int serverBufferSize,
	      int controlMode)
{
  int childpid, pipeData[2], pipeCtrl[2], pipeAck[2];
  
  if (pipe(pipeData) < 0) {
    fprintf(stderr, "error: could not create data pipe\n");
    exit(1);
  }

  if (pipe(pipeCtrl) < 0) {
    fprintf(stderr, "error: could not create control pipe\n");
    exit(1);
  }

  if (pipe(pipeAck) < 0) {
    fprintf(stderr, "error: could not create control pipe\n");
    exit(1);
  }

  if ((childpid = fork()) < 0) {
    fprintf(stderr, "error: could not fork\n");
    exit(1);
  }

  if (childpid > 0) {
    close(pipeData[0]);
    close(pipeCtrl[0]);
    close(pipeAck[1]);
    Client(targetFiles, 
	   pipeData[1], pipeCtrl[1], pipeAck[0], 
	   transferTime, clientBufferSize);
    while (wait((int *) 0) != childpid) {
#ifdef SERVER
      fprintf(stdout, "parent: waiting for child...\n");
#endif
    }
    close(pipeData[1]);
    close(pipeCtrl[1]);
    close(pipeAck[0]);
  } else {
    close(pipeData[1]);
    close(pipeCtrl[1]);
    close(pipeAck[0]);
    Server(sourceFiles, 
	   pipeData[0], pipeCtrl[0], pipeAck[1], 
	   bandwidth, transferTime, transferMode, serverBufferSize, 
	   controlMode);
    close(pipeData[0]);
    close(pipeCtrl[0]);
    close(pipeAck[1]);
  }
}

int 
main(int argc, char **argv) 
{
  char sourceFiles[MAX_FILE_NUM][MAX_FILE_NAME];
  char targetFiles[MAX_FILE_NUM][MAX_FILE_NAME];
  long bandwidth;
  long transferTime; /* long for compatibility with struct timeval */
  int transferMode;
  int clientBufferSize;
  int serverBufferSize;
  int controlMode;

  if (argc < 7) {
    fprintf(stderr, 
	    "usage: %s bandwidth xTime xMode clientBuf serverBuf ctrlMode\n", 
	    argv[0]);
    fflush(stderr);
    exit(1);
  } else {
    bandwidth = atoi(argv[1]);
    transferTime = atoi(argv[2]);
    transferMode = atoi(argv[3]);
    clientBufferSize = atoi(argv[4]);
    serverBufferSize = atoi(argv[5]);
    controlMode = atoi(argv[6]);
  }
  if (transferMode < TM_FIRST || transferMode > TM_LAST) {
    fprintf(stderr, "error: wrong value for transferMode\n");
    exit(1);
  }
  if (controlMode < CM_FIRST || controlMode > CM_LAST) {
    fprintf(stderr, "error: wrong value for controlMode\n");
    exit(1);
  }
  if (serverBufferSize > clientBufferSize) {
    fprintf(stderr, "error: client buffer is smaller than server buffer\n");
    fflush(stderr);
    exit(1);
  }
  if (serverBufferSize > MAX_BUFF || serverBufferSize < MIN_BUFF) {
    fprintf(stderr, "error: wrong value for server buffer\n");
    fflush(stderr);
    exit(1);
  }
  if (clientBufferSize > MAX_BUFF || clientBufferSize < MIN_BUFF) {
    fprintf(stderr, "error: wrong value for client buffer\n");
    fflush(stderr);
    exit(1);
  }
#ifdef SERVER
  fprintf(stdout, "main: bandwidth = %ld\n", bandwidth);
  fprintf(stdout, "main: transfer time = %ld\n", transferTime);
  fprintf(stdout, "main: transfer mode = %d\n", transferMode);
  fprintf(stdout, "main: client buffer size = %d\n", clientBufferSize);
  fprintf(stdout, "main: server buffer size = %d\n", serverBufferSize);
  fflush(stdout);
#endif

  strcpy(sourceFiles[0], 
	 "/afs/cs.wisc.edu/p/qclic/ims.2/RPI/Mobile/mobile000/mobile000.jpg");
/*	 "/afs/cs.wisc.edu/p/qclic/ims.2/USC/couple/couple.jpg");*/
/*	 "/afs/cs.wisc.edu/p/qclic/ims.2/RPI/Tennis/tt001/tt001.jpg");*/
/*	 "/afs/cs.wisc.edu/p/qclic/ims.2/orl_faces/s1/1.pgm/1.pgm.jpg\0");*/
  strcpy(sourceFiles[1], 
	 "/afs/cs.wisc.edu/p/qclic/ims.2/RPI/Mobile/mobile001/mobile001.jpg");
/*	 "/afs/cs.wisc.edu/p/qclic/ims.2/USC/girl/girl.jpg");*/
/*	 "/afs/cs.wisc.edu/p/qclic/ims.2/RPI/Tennis/tt002/tt002.jpg");*/
/*	 "/afs/cs.wisc.edu/p/qclic/ims.2/orl_faces/s1/2.pgm/2.pgm.jpg\0");*/
  strcpy(sourceFiles[2], 
	 "/afs/cs.wisc.edu/p/qclic/ims.2/RPI/Mobile/mobile002/mobile002.jpg");
/*	 "/afs/cs.wisc.edu/p/qclic/ims.2/USC/girl.2/girl.2.jpg");*/
/*	 "/afs/cs.wisc.edu/p/qclic/ims.2/RPI/Tennis/tt003/tt003.jpg");*/
/*	 "/afs/cs.wisc.edu/p/qclic/ims.2/orl_faces/s1/3.pgm/3.pgm.jpg\0");*/
  strcpy(sourceFiles[3], 
	 "/afs/cs.wisc.edu/p/qclic/ims.2/RPI/Mobile/mobile003/mobile003.jpg");
/*	 "/afs/cs.wisc.edu/p/qclic/ims.2/USC/girl3/girl3.jpg");*/
/*	 "/afs/cs.wisc.edu/p/qclic/ims.2/RPI/Tennis/tt004/tt004.jpg");*/
/*	 "/afs/cs.wisc.edu/p/qclic/ims.2/orl_faces/s1/4.pgm/4.pgm.jpg\0");*/
  strcpy(sourceFiles[4], 
	 "/afs/cs.wisc.edu/p/qclic/ims.2/RPI/Mobile/mobile004/mobile004.jpg");
/*	 "/afs/cs.wisc.edu/p/qclic/ims.2/USC/house/house.jpg");*/
/*	 "/afs/cs.wisc.edu/p/qclic/ims.2/RPI/Tennis/tt005/tt005.jpg");*/
/*	 "/afs/cs.wisc.edu/p/qclic/ims.2/orl_faces/s1/5.pgm/5.pgm.jpg\0");*/
  sourceFiles[5][0] = '\0';

  strcpy(targetFiles[0], "1.pgm.jpg\0");
  strcpy(targetFiles[1], "2.pgm.jpg\0");
  strcpy(targetFiles[2], "3.pgm.jpg\0");
  strcpy(targetFiles[3], "4.pgm.jpg\0");
  strcpy(targetFiles[4], "5.pgm.jpg\0");
  targetFiles[5][0] = '\0';

  GetQclicFiles(sourceFiles, 
		targetFiles, 
		bandwidth, 
		transferTime, 
		transferMode,
		clientBufferSize,
		serverBufferSize,
		controlMode);
  
  return 0;
}


