/* gqcf.h */

/* Krzysztof Zmudzinski
   CS 737 Class Project
   Controlled Transfer of Qclic Files
   Date of completion: 5/5/97

   The purpose of this project was to develop an algorithm for
   controlled transfer of Qclic files.  `Controlled' means that the
   user specifies what files should be transferred, how much bandwidth
   should be used to transfer these files, and how much time should
   the transfer take.  The algorithm is responsible for transferring
   the specified files without exceeding the specified bandwidth and time.

   For more info see gqcf.c and the project report at 
   http://www.cs.wisc.edu/~zmudzin/737/737.html
*/

#include <stdio.h>
#include <strings.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/wait.h>
#include <signal.h>
#include <errno.h>
#include <sys/resource.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#define SERVER_PORT1 20000
#define SERVER_PORT2 20001

/* 
   Types of "packets" transferred between client and server:
   (each box represents a long (32-bit) value)
*/

/*
   Transfer request format (client->server):
   +--------------+-----------+---------------+---------------+
   | TRANSFER_REQ | bandwidth | transfer time | transfer mode |
   +--------------+-----------+---------------+---------------+
*/

#define TRANSFER_REQ 1
/* transfer modes */
#define TM_ALT_EXACT 1
#define TM_ALT_NEXT_SCAN 2
#define TM_ALT_PREV_SCAN 3
#define TM_SEQ_EXACT 4
#define TM_SEQ_NEXT_SCAN 5
#define TM_SEQ_PREV_SCAN 6

#define TM_FIRST 1
#define TM_LAST 6

/*
   File request format (client->server):
   (length is length of file name)
   +----------+------+-------+--------+----...----+
   | FILE_REQ | type | file# | length | file name |
   +----------+------+-------+--------+----...----+
*/

#define FILE_REQ 2
/* request types for file request */
#define DEFAULT 1
#define LAST_FILE 2

/*
   File buffer format (server->client):
   (length is length of file contents)
   +----------+------+--------+--------+------...------+
   | FILE_BUF | type |  file# | length | file contents |
   +----------+------+--------+--------+------...------+
*/

#define FILE_BUF 3
/* request types for write file buffer request */
#define CREATE 1
#define APPEND 2

/*
   Transfer summary (server->client):
   +---------------+-------------------+---------------+----------------+
   | TRANSFER_QUIT | bytes transferred | transfer secs | transfer usecs |
   +---------------+-------------------+---------------+----------------+
*/

#define TRANSFER_QUIT 4

#define MAX_FILE_NUM 256
#define MAX_FILE_NAME 256
#define MAX_LINE 80
#define MAX_SCANS 100
#define MILLION 1000000.0
#define MAX_BUFF 4096
#define MIN_BUFF 16

/* control modes */
#define CM_TIMEOUT 1
#define CM_BUFFER 2

#define CM_FIRST 1
#define CM_LAST 2

#ifdef ALL_OUTPUT
#define INFO
#define CTRL
#define FILEOP
#define ACK
#define FILEIO
#define CLIENT
#define SERVER
#define CLIENT
#define WATOUT
#define WAIT
#endif

/* longPtr is used to transfer longs in network byte order using htonl;
   it is allocated somewhere in the file that uses it (client.c, server.c) */

#define LONGSIZE (sizeof(long))
long *longPtr;

/* ReadBytes
   
   reads n bytes from socket described by fd into buffer ptr;
   returns number of bytes read or -1 for error;
*/

int
ReadBytes(int fd,
	  char *ptr,
	  int nbytes)
{
  int nleft, nread;
  
  nleft = nbytes;
  while (nleft > 0) {
    nread = read(fd, ptr, nleft);
    if (nread < 0) {
      return nread;            /* error, return < 0 */
    } else if (nread == 0) {
      break;                   /* EOF */
    }
    nleft -= nread;
    ptr += nread;
  }
  return (nbytes - nleft);          /* return >= 0 */
}

/* WriteBytes

   writes n bytes to a socket described by fd from buffer ptr;
   returns number of bytes written or 0 for error;
*/

int
WriteBytes(int fd,
	   char *ptr,
	   int nbytes)
{
  int nleft, nwritten;

  nleft = nbytes;
  while (nleft > 0) {
    nwritten = write(fd, ptr, nleft);
    if (nwritten <= 0) {
      return nwritten;       /* error */
    }
    nleft -= nwritten;
    ptr += nwritten;
  }
  return (nbytes - nleft);   /* it's really nbytes */
}

/* ReadLine
   
   reads one line from file described by iFD into buffer sBuff;
   replaces '\n' with '\0';
   returns number of characters read or EOF
*/

int 
ReadLine(char sBuff[], 
	 int iFD)
{
  int iCount;
  char cChar;

  iCount = 0;
  
  while (read(iFD, &cChar, 1) == 1) {
    if (cChar == '\n') {
      sBuff[iCount] = '\0';
      return (iCount + 1);
    }
    sBuff[iCount] = cChar;
    iCount++;
  }
  
  sBuff[iCount] = '\0';
  return EOF;
}

int
SendHeader(int sockfd,
	   long headerType, 
	   long long1,
	   long long2,
	   long long3)
{
  char buff[4 * LONGSIZE];

  *longPtr = htonl(headerType);
  memcpy(&buff, longPtr, LONGSIZE);

  *longPtr = htonl(long1);
  memcpy(&buff[LONGSIZE], longPtr, LONGSIZE);

  *longPtr = htonl(long2);
  memcpy(&buff[2 * LONGSIZE], longPtr, LONGSIZE);

  *longPtr = htonl(long3);
  memcpy(&buff[3 * LONGSIZE], longPtr, LONGSIZE);

  if (WriteBytes(sockfd, buff, sizeof(buff)) != sizeof(buff)) {
    return -1;
  }

  return 0;
}

int
ReadHeader(int sockfd,
	   long *headerType,
	   long *long1,
	   long *long2,
	   long *long3)
{
  char buff[4 * LONGSIZE];

  if (ReadBytes(sockfd, buff, sizeof(buff)) != sizeof(buff)) {
    return -1;
  }

  memcpy(longPtr, &buff, LONGSIZE);
  *headerType = ntohl(*longPtr);

  memcpy(longPtr, &buff[LONGSIZE], LONGSIZE);
  *long1 = ntohl(*longPtr);

  memcpy(longPtr, &buff[2 * LONGSIZE], LONGSIZE);
  *long2 = ntohl(*longPtr);

  memcpy(longPtr, &buff[3 * LONGSIZE], LONGSIZE);
  *long3 = ntohl(*longPtr);

  return 0;
}

void
sig_child()
{
#ifdef BSD
  int pid;
  union wait status;

  while ((pid = wait3(&status, WNOHANG, (struct rusage *) 0)) > 0);
#endif
}
