/* server.c */

/* Krzysztof Zmudzinski
   CS 737 Class Project
   Controlled Transfer of Qclic Files
   Date of completion: 5/5/97

   The purpose of this project was to develop an algorithm for
   controlled transfer of Qclic files.  `Controlled' means that the
   user specifies what files should be transferred, how much bandwidth
   should be used to transfer these files, and how much time should
   the transfer take.  The algorithm is responsible for transferring
   the specified files without exceeding the specified bandwidth and time.

   For more info see gqcf.c and the project report at
   http://www.cs.wisc.edu/~zmudzin/737/737.html
*/

#include "gqcf.h"

double minwait = 0, maxwait = 0, totalwait = 0;
int numofwaits = 0;
int minbuf = 0, maxbuf = 0;

/* ReadFileInfo

   opens jsdir file associated with fileName, reads information
   necessary to transfer fileName, and returns it in passed arguments;

   returns 0 if jsdir file exists and contains information for at
   last two layers; returns -1 otherwise;

   format of jsdir file:
   ---------------------
   scan offset <offset>
   scan offset <offset>
   ...
   end offset <offset>
   #if nplanes==1
   PSNR: <psnr> dB
   PSNR: <psnr> dB
   ..
   #elseif nplanes==3
   RPSNR: <rpsnr> dB GPSNR: <gpsnr> dB BPSNR: <bpsnr> dB
   or
   YPSNR: <rpsnr> dB CbPSNR: <gpsnr> dB CrPSNR: <bpsnr> dB
   ...
  
   only number of scans (layers) and corresponding offsets are used;
   psnr, gpsnr, and bpsnr may be used in future modifications;
*/

int 
ReadFileInfo(char *fileName,
	     int *nscansplus1,
	     int *offsets, 
	     double *psnr, 
	     double *gpsnr, 
	     double *bpsnr)
{

  char jsdirname[MAX_FILE_NAME];
  int fd, l, i, from;
  char buff[MAX_LINE];
  char foo1[20], foo2[20], foo3[20], foo4[20], foo5[20], foo6[20];
  
  sprintf(jsdirname, "%s.jsdir", fileName);
  if ((fd = open(jsdirname, O_RDONLY)) < 0) {
    fprintf(stderr, "server error: could not open %s\n", jsdirname);
    fflush(stderr);
    return -1;
  }
      
  *nscansplus1 = 0;
  l = ReadLine(buff, fd);
  while ((l != EOF) && (strncmp(buff, "end", 3))) {
    sscanf(buff, "%s%s%d", foo1, foo2, &offsets[*nscansplus1]);
    (*nscansplus1)++;
    l = ReadLine(buff, fd);
  }
  if (l == EOF) {
    close(fd);
    return -1;
  }
  sscanf(buff, "%s%s%d", foo1, foo2, &offsets[*nscansplus1]);
  (*nscansplus1)++;

  /* if only header or incomplete information present */
  if (*nscansplus1 < 2) {
    return -1;
  }
  
  i = 1;
  l = ReadLine(buff, fd);
  from = 0;
  
  while ((l != EOF) && (i < *nscansplus1)) {
    if (!strncmp(buff, "RPSNR", 5)) {
      sscanf(buff, "%s%lf%s%s%lf%s%s%lf%s", 
	     foo1, &psnr[i], foo2, foo3, &gpsnr[i], 
	     foo4, foo5, &bpsnr[i], foo6);

#ifdef INFO
      fprintf(stdout, "server: size:%d B, R:%f dB, G:%f dB, B:%f dB\n",
	      offsets[i] - from, psnr[i], gpsnr[i], bpsnr[i]);
      fflush(stdout);
#endif

    } else if (!strncmp(buff,"YPSNR", 5)) {
      sscanf(buff, "%s%lf%s%s%lf%s%s%lf%s", 
	     foo1, &psnr[i], foo2, foo3, &gpsnr[i], 
	     foo4, foo5, &bpsnr[i], foo6);

#ifdef INFO
      fprintf(stdout, "server: size:%d B, Y:%f dB, Cb:%f dB, Cr:%f dB\n",
	      offsets[i] - from, psnr[i], gpsnr[i], bpsnr[i]);
      fflush(stdout);
#endif

    } else {
      sscanf(buff, "%s%lf%s", foo1, &psnr[i], foo2);

#ifdef INFO
      fprintf(stdout, "server: size:%d B, Gray:%f dB\n", 
	      offsets[i] - from, psnr[i]);
      fflush(stdout);
#endif

    }
    from = offsets[i];
    i++;
    l = ReadLine(buff, fd);
  }

  close(fd);
  
#ifdef INFO
  if (i < *nscansplus1)  {
    while (i < *nscansplus1) {
      fprintf(stdout, "server: size:%d B\n", offsets[i] - from);
      fflush(stdout);
      from = offsets[i];
      i++;
    }
    fprintf(stdout, "END\n");
    fflush(stdout);
  }

  fprintf(stdout, "END\n");
  fflush(stdout);
#endif

  return 0;
  
}

/* WaitAfter

   controls transfer of files by delaying program execution;
   there are two ways in which program can be dalayed:
   1. setting timeout;
   2. changing the size of the buffer used to transfer files;

   only option 1. works well but it allows for bursts;
   option 2. is supposed to eliminate bursts and force
   transfer into a  steady data streaming; at this point
   this option does not work very well - buffer size varies
   a lot during the life of a connection;

*/

void
WaitAfter(long totalSize, 
	  long bandwidth,
	  int *bufferSize,
	  double usecStart,
	  int controlMode)
{
  struct timeval tcurr, twait;
  double usecCurr, usec;

  if (gettimeofday(&tcurr, (struct timezone *) 0) < 0) {
    return;
  } else {
    usecCurr = (double) tcurr.tv_sec * MILLION + (double) tcurr.tv_usec;
  }

  if (controlMode == CM_BUFFER) {
    *bufferSize = ((usecCurr - usecStart) * (double) bandwidth * 
		   (double) *bufferSize) / (MILLION * (double) totalSize);

    if (*bufferSize < MIN_BUFF) {
      *bufferSize = MIN_BUFF;
    } else if (*bufferSize > MAX_BUFF) {
      *bufferSize = MAX_BUFF;
    }

#ifdef WATOUT
    fprintf(stdout, "wait: new buffer: %d\n", *bufferSize);
    fflush(stdout);
#endif

    if (*bufferSize < minbuf) {
      minbuf = *bufferSize;
    } else if (*bufferSize > maxbuf) {
      maxbuf = *bufferSize;
    }

  } else {

    /* calculate delay */
    usec = (MILLION * (double) totalSize) / (double) bandwidth + usecStart
      - usecCurr;
    
#ifdef WATOUT
    fprintf(stdout, "wait: usec: %f\n", usec);
    fflush(stdout);
#endif
    
    /* instead of 0, some threshold could be used to avoid small
       waits; this could result in bigger bursts, however */
    if (usec < 0) {
      return;
    }

    /* some stats */
    if (usec < minwait) {
      minwait = usec;
    } else if (usec > maxwait) {
      maxwait = usec;
    }
    totalwait += usec;
    numofwaits++;
    twait.tv_sec = (long) (usec / MILLION);
    twait.tv_usec = (long) (usec - (double) twait.tv_sec * MILLION);
    
#ifdef WATOUT
    fprintf(stdout, "wait: twait.tv_sec: %ld; twait.tv_usec: %ld\n", 
	    twait.tv_sec, twait.tv_usec);
    fflush(stdout);
#endif
    
    /* wait */
    if (select(0, 
	       (fd_set *) 0, 
	       (fd_set *) 0, 
	       (fd_set *) 0, 
	       &twait) < 0) {
      fprintf(stderr, "server error: select error\n");
      fflush(stderr);
    }
  }
}

/* TransferFile

   opens fileName, seeks past its last read byte (from), reads nbytes,
   and transfers nbytes into bufferSizes chunks;

   waits after each transfer (see WaitAfter);
   
   returns -1 for error; 0 otherwise;

*/

int
TransferFile(char *fileName,
	     long fileNum,
	     int from,
	     int nbytes,
	     int sockfd,
	     int *bufferSize, 
	     int *dataRead,
	     long *totalSize,
	     int *currSize,
	     long transferMode, 
	     long bandwidth,
	     long maxSize,
	     double usecStart,
	     int controlMode)
{
  char buff[MAX_BUFF];
  int nloops, remaining;
  int n, fd;
  long reqType;
  
  if ((fd = open(fileName, O_RDONLY, 0)) < 0) {
    fprintf(stderr, "server error: could not open %s\n", fileName);
    fflush(stderr);
    return -1;
  }
  
  if (lseek(fd, from, SEEK_SET) < 0) {
    fprintf(stderr, "server error: seek error\n");
    fflush(stderr);
    close(fd);
    return -1;
  }

  nloops = nbytes / *bufferSize;
  remaining = nbytes % *bufferSize;
  
  while (nloops > 0) {
    if ((n = read(fd, buff, *bufferSize)) < 0) {
      fprintf(stderr, "server error: read error\n");
      fflush(stderr);
      close(fd);
      return -1;
    }
    
    /* we have read some data */
    *dataRead = 1;

#ifdef FILEIO
    fprintf(stdout, "server: %d bytes read\n", n);
    fflush(stdout);
#endif

    /* if this is first of the current file, let the client know
       that it should create it, otherwise open for append */
    if (*currSize == 0) {
      reqType = CREATE;
    } else {
      reqType = APPEND;
    }
    if (SendHeader(sockfd, FILE_BUF, reqType, fileNum, *bufferSize) < 0) {
      close(fd);
      return -1;
    }
    if ((n = WriteBytes(sockfd, buff, *bufferSize)) < 0) {
      fprintf(stderr, "server error: send error\n");
      fflush(stderr);
      close(fd);
      return -1;
    }
    
    *totalSize += n;
    *currSize += n;

#ifdef FILEIO
    fprintf(stdout, "server: %d bytes written\n", n);
    fflush(stdout);
#endif

    if (transferMode == TM_ALT_EXACT || transferMode == TM_ALT_PREV_SCAN) {
      if (*totalSize >= maxSize) {
	break;
      }
    }

    if (transferMode == TM_SEQ_EXACT || transferMode == TM_SEQ_PREV_SCAN) {
      if (*currSize >= maxSize) {
	break;
      }
    }

#ifdef WAIT
    WaitAfter(*totalSize, bandwidth, bufferSize, usecStart, controlMode);
#endif

    if (controlMode == CM_TIMEOUT) {
      nloops--;
    } else {
      nbytes -= n;
      nloops = nbytes / *bufferSize;
      remaining = nbytes % *bufferSize;
    }
  
  }
  
  if (remaining < 1) {
    close(fd);
    return 0;
  }

  if ((*totalSize < maxSize || transferMode == TM_ALT_NEXT_SCAN) ||
      (*currSize < maxSize || transferMode == TM_SEQ_NEXT_SCAN)) {
    if ((n = read(fd, buff, remaining)) < 0) {
      fprintf(stderr, "server error: read error\n");
      fflush(stderr);
      close(fd);
      return -1;
    }
    
    *dataRead = 1;

#ifdef FILEIO
    fprintf(stdout, "server: %d bytes read\n", n);
    fflush(stdout);
#endif

    if (*currSize == 0) {
      reqType = CREATE;
    } else {
      reqType = APPEND;
    }
    if (SendHeader(sockfd, FILE_BUF, reqType, fileNum, remaining) < 0) {
      close(fd);
      return -1;
    }
    if ((n = WriteBytes(sockfd, buff, remaining)) < 0) {
      fprintf(stderr, "server error: send error\n");
      fflush(stderr);
      close(fd);
      return -1;
    }
      
    *totalSize += n;
    *currSize += n;

#ifdef FILEIO
    fprintf(stdout, "server: %d bytes written\n", n);
    fflush(stdout);
#endif

#ifdef WAIT
    WaitAfter(*totalSize, bandwidth, bufferSize, usecStart, controlMode);
#endif
  }

  close(fd);
  return 0;

}

/* GetScanExtent

   opens fileName and based on currScan, passes back scan's start
   position (from) and its size (toread);

   returns -1 for error; 1 if EOF; 0 otherwise;
*/

int
GetScanExtent(char *fileName,
	      int currScan,
	      int *from,
	      int *toread)
{
  int offsets[MAX_SCANS];
  int numscans;

  /* hook for quality */
  double psnr[MAX_SCANS], gpsnr[MAX_SCANS], bpsnr[MAX_SCANS];

  if (ReadFileInfo(fileName, &numscans, offsets, psnr, gpsnr, bpsnr) != 0) {
    fprintf(stderr, "server error: error reading %s.jsdir\n", fileName);
    fflush(stderr);
    return -1;
  }

  if (currScan > numscans - 1) {
    return 1;
  }
  
  /* there is an error in the original algorithm (found in jview.c):
     1. an image file starts at byte 0 instead of 1;
     2. the beginning of a scan is used as the size of the previous 
     scan which is 1 byte bigger than the actual scan size since the 
     file starts at 0;
     3. the end of the last scan is used as such 
  */
  
  if (currScan == 1) {
    *from = 1; /* 1. */
  } else {
    *from = offsets[currScan - 1];
  }
  
  *toread = offsets[currScan] - *from; /* 1. */
  /* last offset is the end of the offset, not the
     beginning of the next one (2. & 3.) */
  if (currScan == numscans - 1) {
    (*toread)++;
  }

#ifdef INFO
  fprintf(stdout, "server: scan # %d starts at %d and has %d bytes\n", 
	  currScan, *from, *toread);
  fflush(stdout);
#endif
      
  return 0;
  
}

/* TransferFileAlt

   transfers files alternating between them, transferring one scan
   of each file until transferSize is reached;

   returns actual number of bytes transferred;
*/

int
TransferFileAlt(char sourceFiles[][MAX_FILE_NAME],
		int sockfd,
		long bandwidth, 
		long transferSize,
		int transferMode,
		int bufferSize,
		double usecStart,
		int controlMode)
{
  int fn;
  int from, toread;
  long totalSize;
  int currSize, currScan;
  int dataRead;

  totalSize = 0;

  currScan = 0;
  /* assume that some data will be read */
  dataRead = 1;

  while ((totalSize < transferSize) && dataRead) {

    currScan++;
    
    fn = 0;
    /* assume that no data will be read */
    dataRead = 0;
    
    while (sourceFiles[fn][0] != '\0') {
      
      /* needed by TransferFile to let client know when to
	 create a file and when to append to it */
      if (currScan == 1) {
	currSize = 0;
      }
      if (GetScanExtent(sourceFiles[fn], currScan, &from, &toread) != 0) {
	fn++;
	continue;
      }

      if (totalSize + toread > transferSize) {
	if (transferMode == TM_ALT_EXACT) {
	  toread = transferSize - totalSize;
	} else if (transferMode == TM_ALT_PREV_SCAN) {
	  break;
	}
      }

      if (TransferFile(sourceFiles[fn],
		       fn,
		       from - 1, 
		       toread,
		       sockfd,
		       &bufferSize, 
		       &dataRead,
		       &totalSize,
		       &currSize,
		       transferMode,
		       bandwidth,
		       transferSize,
		       usecStart,
		       controlMode) != 0) {
	fprintf(stderr, "server error: transfer file failed\n");
	fflush(stderr);
      }

      fn++;
      
      if (totalSize >= transferSize) {
	break;
      }

    }
    
    if (transferMode == TM_ALT_PREV_SCAN) {
      if (totalSize + toread > transferSize) {
	break;
      }
    }
    
  }
  
  return totalSize;

}

/* TransferFileSeq

   transfers files sequentially dividing transferSize equally aong
   all files;

   returns the actual number of bytes transferred;
*/

int
TransferFileSeq(char sourceFiles[][MAX_FILE_NAME],
		int sockfd,
		long bandwidth, 
		long transferSize,
		int transferMode,
		int bufferSize,
		double usecStart,
		int controlMode)
{
  int numberOfFiles, fn;
  int from, toread;
  long shareSize, totalSize;
  int currSize, currScan;
  int dataRead;

  /* find number of files */
  fn = 0;
  while (sourceFiles[fn][0] != '\0') {
    fn++;
  }
  
  /* calculate each file's share of transferSize */
  numberOfFiles = fn;
  shareSize = transferSize / numberOfFiles;

  totalSize = 0;

  for (fn = 0; fn < numberOfFiles; fn++) {
    
    currScan = 0;
    currSize = 0;

    dataRead = 1;
    
    while ((currSize < shareSize) && dataRead) {
      
      currScan++;
      dataRead = 0;
      
      if (GetScanExtent(sourceFiles[fn], currScan, &from, &toread) != 0) {
	break;
      }
      
      if (currSize + toread > shareSize) {
	if (transferMode == TM_SEQ_EXACT) {
	  toread = shareSize - currSize;
	} else if (transferMode == TM_SEQ_PREV_SCAN) {
	  break;
	}
      }  

      if (TransferFile(sourceFiles[fn],
		       fn,
		       from - 1, 
		       toread,
		       sockfd,
		       &bufferSize, 
		       &dataRead,
		       &totalSize,
		       &currSize,
		       transferMode,
		       bandwidth,
		       shareSize,
		       usecStart,
		       controlMode) != 0) {
	fprintf(stderr, "server error: transfer file failed\n");
	fflush(stderr);
      }
      
    }
    
  }

  return totalSize;
  
}

/* ReadTransferInfo

   reads transfer request from client;

   returns -1 for error; 0 otherwise
*/

int
ReadTransferInfo(int sockfd,
		 char sourceFiles[][MAX_FILE_NAME],
		 long *bandwidth,
		 long *transferTime,
		 long *transferMode)
{
  long headerType;
  long reqType;
  long fileNum;
  long nameLen;

  if (ReadHeader(sockfd, &headerType, 
		 bandwidth, transferTime, transferMode) < 0) {
    return -1;
  }

  if (headerType != TRANSFER_REQ) {
    fprintf(stderr, "server error: wrong header type\n");
    fflush(stderr);
    return -1;
  }

  fprintf(stdout, "bandwidth:%ld; transferTime:%ld; transferMode:%ld\n",
	  *bandwidth, *transferTime, *transferMode);
  fflush(stdout);

  while (1) {
    if (ReadHeader(sockfd, &headerType, &reqType, &fileNum, &nameLen) < 0) {
      return -1;
    }
    if (headerType != FILE_REQ) {
      fprintf(stderr, "server error: wrong header type\n");
      fflush(stderr);
      return -1;
    }
    if (ReadBytes(sockfd, sourceFiles[fileNum], nameLen) < 0) {
      return -1;
    }
    sourceFiles[fileNum][nameLen] = '\0';

    fprintf(stdout, "%s\n", sourceFiles[fileNum]);
    fflush(stdout);

    if (reqType == LAST_FILE) {
      break;
    }
  }
  
  return 0;
}

/* Server

   receives transfer request from client, transfers files according to
   controlMode, terminates transfer and prints out some statistics;
*/

void 
Server(int sockfd,
       int bufferSize,
       int controlMode)
{
  long bandwidth;
  long transferTime;
  long transferMode;
  char sourceFiles[MAX_FILE_NUM][MAX_FILE_NAME];

  struct timeval tstart, tend;
  double usecStart, usecEnd;
  long transferSize;

  /* used by memcpy in various places to transfer 
     longs in network byte order (htonl) */
  longPtr = (long *) malloc(LONGSIZE);

  if (ReadTransferInfo(sockfd, 
		       sourceFiles, 
		       &bandwidth, &transferTime, &transferMode) < 0) {
    fprintf(stderr, "server error: read transfer info error\n");
    fflush(stderr);
    return;
  }

  if (gettimeofday(&tstart, (struct timezone *) 0) < 0) {
    fprintf(stderr, "server error: gettimeofday failed\n");
  }

  usecStart = (double) tstart.tv_sec * MILLION + (double) tstart.tv_usec;

  transferSize = bandwidth * transferTime;

  switch (transferMode) {
  case TM_ALT_EXACT :
  case TM_ALT_NEXT_SCAN :
  case TM_ALT_PREV_SCAN :
    transferSize = TransferFileAlt(sourceFiles, sockfd,
				   bandwidth, transferSize, transferMode, 
				   bufferSize, usecStart, controlMode);
    break;
  case TM_SEQ_EXACT :
  case TM_SEQ_NEXT_SCAN :
  case TM_SEQ_PREV_SCAN :
    transferSize = TransferFileSeq(sourceFiles, sockfd,
				   bandwidth, transferSize, transferMode, 
				   bufferSize, usecStart, controlMode);
    break;
  }

  if (gettimeofday(&tend, (struct timezone *) 0) < 0) {
    usecEnd = usecStart + transferTime;
  } else {
    usecEnd = (double) tend.tv_sec * MILLION + (double) tend.tv_usec;
  }

  (void) SendHeader(sockfd, 
		    TRANSFER_QUIT, transferSize, tend.tv_sec, tend.tv_usec);

  shutdown(sockfd, 2);

  /* print out some stats */
  fprintf(stdout, 
	  "server: transferred %ld bytes in %f sec at rate of %f Bs\n", 
	  transferSize, (usecEnd - usecStart) / MILLION, 
	  (double) transferSize / ((usecEnd - usecStart) / MILLION));
  fprintf(stdout, "minwait:%f;maxwait:%f;totalwait:%f;numofwaits:%d\n", 
	  minwait, maxwait, totalwait, numofwaits);
  fprintf(stdout, "minbuf:%d;maxbuf:%d\n", minbuf, maxbuf);
  fflush(stdout);
  
}

/* SendQclicFiles

   opens port and listens, if there is a connection requests, it accepts
   up to 5 different requests and forks new process that runs Server to
   process client requests;
*/

void
SendQclicFiles(int bufferSize,
	       int controlMode)
{

  int sockfd, newsockfd, cli_len;
  struct sockaddr_in cli_addr, serv_addr;
  int childpid;

  sockfd = socket(AF_INET, SOCK_STREAM, 0);
  if (sockfd < 0) {
    perror("server error: cannot open socket");
    exit(1);
  }

  bzero((char *) &serv_addr, sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;
  serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  serv_addr.sin_port = htons(SERVER_PORT1);

  /* try two different ports if one is busy */
  if (bind(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0) {
    serv_addr.sin_port = htons(SERVER_PORT2);
    if (bind(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0) {
      perror("server error: cannot bind local address");
      exit(1);
    }
  }

  listen(sockfd, 5);

  while (1) {
  
    cli_len = sizeof((struct sockaddr *) &cli_addr);
    newsockfd = accept(sockfd, (struct sockaddr *) &cli_addr, &cli_len);
    
    if (newsockfd < 0) {
      perror("server error: accept failed");
      continue;
    }

    if ((childpid = fork()) < 0) {
      fprintf(stderr, "server error: could not fork\n");
      continue;
    }

    if (childpid == 0) {
      close(sockfd);
      Server(newsockfd, bufferSize, controlMode);
      exit(0);
    }

    close(newsockfd);

  }

  close(sockfd);

}  


int 
main(int argc, char **argv) 
{
  int bufferSize;
  int controlMode;

  if (argc != 3) {
    fprintf(stderr, "usage: %s BufferSize ControlMode\n", argv[0]);
    fflush(stderr);
    exit(1);
  } else {
    bufferSize = atoi(argv[1]);
    controlMode = atoi(argv[2]);
  }

  if (bufferSize > MAX_BUFF || bufferSize < MIN_BUFF) {
    fprintf(stderr, "server error: wrong value for server buffer\n");
    fflush(stderr);
    exit(1);
  }

  if (controlMode > CM_LAST || controlMode < CM_FIRST) {
    fprintf(stderr, "server error: wrong control type\n");
    fflush(stderr);
    exit(1);
  }

#ifdef SERVER
  fprintf(stdout, "server: server buffer size = %d\n", bufferSize);
  fprintf(stdout, "server: control mode = %d\n", controlMode);
  fflush(stdout);
#endif

  /* we are not interested in the exit status of our children
     so will will not wait for them; we don't want them to
     become zombies either; with System V we can just ignore
     the signal; with BSD we have to catch the signal and 
     execute the wait3 system call */

#ifdef BSD
printf("here\n");
  signal(SIGCLD, sig_child); /* BSD */
#else
  signal(SIGCLD, SIG_IGN);   /* System V */
#endif

  SendQclicFiles(bufferSize, controlMode);

  return 0;
}
